-- Adjust this setting to control where the objects get created.

-- complain if script is sourced in psql, rather than via CREATE EXTENSION
\echo Use "CREATE EXTENSION sqlserver_compatible" to load this file. \quit

SET search_path = public;


--- <CHARINDEX> ------------------------------------------------
CREATE OR REPLACE FUNCTION charindex(varchar,varchar)
RETURNS integer
AS $$ SELECT position($1 IN $2); $$
LANGUAGE SQL IMMUTABLE STRICT;

----------------------------------------------------------------
CREATE OR REPLACE FUNCTION charindex(varchar,varchar,numeric) 
RETURNS integer 
AS $$ 
DECLARE
    rcnt integer;
    wkstart ALIAS FOR $3;
BEGIN
    IF wkstart < 1 THEN
        wkstart = 1;
    END IF;
SELECT position($1 IN substring($2, CAST(trunc(wkstart) AS integer))) INTO rcnt;
    IF rcnt > 0 THEN
        RETURN rcnt+trunc(wkstart)-1;
    END IF;
RETURN 0;
END; $$ 
LANGUAGE PLPGSQL IMMUTABLE STRICT;


--- <TEXT+TEXT BYTEA+BYTEA> ------------------------------------
CREATE OPERATOR + (
   leftarg = text
   ,rightarg = text
   ,procedure = textcat
   ,commutator = +
);

----------------------------------------------------------------
CREATE OPERATOR + (
	leftarg = bytea,
    rightarg = bytea,
    procedure = byteacat,
    commutator = +
);


--- <SQUARE> ---------------------------------------------------
CREATE OR REPLACE FUNCTION pg_catalog.square(double precision)
RETURNS double precision
AS $$ SELECT pg_catalog.power($1, 2); $$
LANGUAGE SQL IMMUTABLE STRICT;

----------------------------------------------------------------
CREATE OR REPLACE FUNCTION pg_catalog.square(money)
RETURNS double precision
AS $$ SELECT pg_catalog.power(CAST(CAST($1 AS numeric) AS double precision) ,2); $$
LANGUAGE SQL IMMUTABLE STRICT;


--- <LEN> ------------------------------------------------------
CREATE OR REPLACE FUNCTION pg_catalog.len(text)
RETURNS integer
AS $$ SELECT pg_catalog.char_length(pg_catalog.rtrim($1)); $$
LANGUAGE SQL IMMUTABLE STRICT;

----------------------------------------------------------------
CREATE OR REPLACE FUNCTION pg_catalog.len(bytea)
RETURNS integer
AS $$ SELECT pg_catalog.char_length(pg_catalog.rtrim(convert_from($1,getdatabaseencoding()))); $$
LANGUAGE SQL IMMUTABLE STRICT;


--- <REPLICATE> -------------------------------------------------
CREATE OR REPLACE FUNCTION replicate(varchar,double precision)
RETURNS varchar
AS $$ 
DECLARE
    retval varchar;
BEGIN
    IF $2 < 0 THEN
        RETURN NULL;
    ELSE
        SELECT repeat($1,CAST(trunc($2) AS integer)) INTO retval;
        RETURN retval;
    END IF;
END; $$ 
LANGUAGE PLPGSQL IMMUTABLE STRICT;

----------------------------------------------------------------
CREATE OR REPLACE FUNCTION replicate(bytea,double precision)
RETURNS varchar
AS $$ 
DECLARE
    retval varchar;
BEGIN
    IF $2 < 0 THEN
        RETURN NULL;
    ELSE
        SELECT repeat(convert_from($1,getdatabaseencoding()),CAST(trunc($2) AS integer)) INTO retval;
        RETURN retval;
    END IF;
END; $$ 
LANGUAGE PLPGSQL IMMUTABLE STRICT;

----------------------------------------------------------------
CREATE OR REPLACE FUNCTION replicate(varchar,money)
RETURNS varchar
AS $$ 
select replicate($1,cast($2 as numeric));
$$ 
LANGUAGE SQL IMMUTABLE STRICT;

----------------------------------------------------------------
CREATE OR REPLACE FUNCTION replicate(bytea,money)
RETURNS varchar
AS $$ 
select replicate($1,cast($2 as numeric));
$$ 
LANGUAGE SQL IMMUTABLE STRICT;


--- <SPACE> -----------------------------------------------------
CREATE OR REPLACE FUNCTION space(double precision)
RETURNS varchar
AS $$ 
DECLARE
	retval varchar;
BEGIN
	IF $1 < 0 THEN
		RETURN NULL;
	ELSE
		SELECT repeat(' ',CAST(trunc($1) AS integer)) INTO retval;
		RETURN retval;
	END IF;
END; $$ 
LANGUAGE PLPGSQL IMMUTABLE STRICT;

----------------------------------------------------------------
CREATE OR REPLACE FUNCTION space(money)
RETURNS varchar
AS $$ SELECT space (cast(cast($1 as numeric) as double precision)); $$
LANGUAGE SQL IMMUTABLE STRICT;


--- <STR> ------------------------------------------------------
CREATE OR REPLACE FUNCTION str(numeric,bigint,bigint)
RETURNS varchar
AS $$
DECLARE
	retval varchar;
	length_integer integer;
	length_point integer;
	point integer;
	decimal_number integer;
	tmp_area numeric;
	length_tmp integer;
BEGIN
	IF $1 IS NULL OR $2 IS NULL OR $2 <= 0 OR $3 < 0 THEN
		RETURN NULL;
	END IF;

	decimal_number = CAST($3 AS integer);
	IF decimal_number IS NULL THEN
		decimal_number = 0;
	ELSIF decimal_number > 16 THEN
		decimal_number = 16;
	END IF;

	--Length of integer part 

	SELECT len(CAST(round($1) AS varchar)) INTO length_integer;

	IF length_integer = $2 THEN
		SELECT CAST(round($1,0) AS varchar) INTO retval;
		RETURN retval;
	ELSIF length_integer > $2 THEN
		SELECT (replicate('*',$2)) INTO retval;
		RETURN retval;
	END IF;

	IF decimal_number = 0 THEN
		SELECT lpad(CAST(round($1,0) AS varchar),CAST($2 AS integer),' ') INTO retval;
		RETURN retval;
	END IF;

	--It asks for the position of a decimal point
	SELECT strpos(CAST($1 AS varchar),'.') INTO length_point;
	point = 1;
	IF length_point = 0 THEN
		point = 0;
	END IF;

	SELECT round($1,decimal_number) INTO tmp_area;
	SELECT len(CAST(tmp_area AS varchar)) INTO length_tmp;
	IF length_tmp <= $2 THEN
		SELECT lpad(CAST(tmp_area AS varchar),CAST($2 AS integer),' ') INTO retval;
	ELSE
		SELECT CAST(round($1, CAST($2 AS integer) - length_integer - point) AS varchar) INTO retval;
	END IF;
	RETURN retval;
END; $$
LANGUAGE PLPGSQL IMMUTABLE CALLED ON NULL INPUT;

----------------------------------------------------------------
CREATE OR REPLACE FUNCTION str(double precision,bigint,bigint)
RETURNS varchar
AS $$ SELECT str(CAST($1 AS numeric), $2, $3); $$
LANGUAGE SQL IMMUTABLE CALLED ON NULL INPUT;

----------------------------------------------------------------
CREATE OR REPLACE FUNCTION str(money,bigint,bigint)
RETURNS varchar
AS $$ SELECT str(CAST($1 AS numeric), $2, $3); $$
LANGUAGE SQL IMMUTABLE CALLED ON NULL INPUT;

----------------------------------------------------------------
CREATE OR REPLACE FUNCTION str(double precision)
RETURNS varchar
AS $$ SELECT str($1, 10, 0); $$
LANGUAGE SQL IMMUTABLE CALLED ON NULL INPUT;

----------------------------------------------------------------
CREATE OR REPLACE FUNCTION str(money)
RETURNS varchar
AS $$ SELECT str($1, 10, 0); $$
LANGUAGE SQL IMMUTABLE CALLED ON NULL INPUT;

----------------------------------------------------------------
CREATE OR REPLACE FUNCTION str(double precision,bigint)
RETURNS varchar
AS $$ SELECT str($1, $2, 0); $$
LANGUAGE SQL IMMUTABLE CALLED ON NULL INPUT;

----------------------------------------------------------------
CREATE OR REPLACE FUNCTION str(money,bigint)
RETURNS varchar
AS $$ SELECT str($1, $2, 0); $$
LANGUAGE SQL IMMUTABLE CALLED ON NULL INPUT;


--<STUFF>------------------------------------------------------
CREATE OR REPLACE FUNCTION stuff(varchar,bigint,bigint,varchar)
RETURNS varchar
AS $$
DECLARE
	retval varchar;
BEGIN
	IF $1 IS NULL OR $2 IS NULL OR $3 IS NULL THEN
		RETURN NULL;
	ELSIF character_length($1) < $2 OR $2 <= 0 OR $3 < 0 THEN
		RETURN NULL;
	ELSIF $4 IS NULL THEN
		SELECT overlay($1 PLACING '' FROM CAST($2 AS integer) FOR CAST($3 AS integer)) INTO retval;
		RETURN retval;
	END IF;

	SELECT overlay($1 PLACING $4 FROM CAST($2 AS integer) FOR CAST($3 AS integer)) INTO retval;
	RETURN retval;
END; $$
LANGUAGE PLPGSQL IMMUTABLE CALLED ON NULL INPUT;

----------------------------------------------------------------
CREATE OR REPLACE FUNCTION stuff(bytea,bigint,bigint,varchar)
RETURNS varchar
AS $$
SELECT stuff(convert_from($1,getdatabaseencoding()),$2,$3,$4);
$$
LANGUAGE SQL IMMUTABLE CALLED ON NULL INPUT;

----------------------------------------------------------------
CREATE OR REPLACE FUNCTION stuff(bytea,bigint,bigint,bytea)
RETURNS varchar
AS $$
SELECT stuff(convert_from($1,getdatabaseencoding()),$2,$3,convert_from($4,getdatabaseencoding()));
$$
LANGUAGE SQL IMMUTABLE CALLED ON NULL INPUT;

----------------------------------------------------------------
CREATE OR REPLACE FUNCTION stuff(varchar,bigint,bigint,bytea)
RETURNS varchar
AS $$
SELECT stuff($1,$2,$3,convert_from($4,getdatabaseencoding()));
$$
LANGUAGE SQL IMMUTABLE CALLED ON NULL INPUT;


--<DATALENGTH>--------------------------------------------------
CREATE OR REPLACE FUNCTION datalength(double precision)
RETURNS integer
AS $$ SELECT 8; $$
LANGUAGE SQL IMMUTABLE STRICT;

----------------------------------------------------------------
CREATE OR REPLACE FUNCTION datalength(real)
RETURNS integer
AS $$ SELECT 4; $$
LANGUAGE SQL IMMUTABLE STRICT;

----------------------------------------------------------------
CREATE OR REPLACE FUNCTION datalength(numeric)
RETURNS integer
AS $$ SELECT pg_column_size($1); $$
LANGUAGE SQL IMMUTABLE STRICT;

----------------------------------------------------------------
CREATE OR REPLACE FUNCTION datalength(bigint)
RETURNS integer
AS $$ SELECT 8; $$
LANGUAGE SQL IMMUTABLE STRICT;

----------------------------------------------------------------
CREATE OR REPLACE FUNCTION datalength(integer)
RETURNS integer
AS $$ SELECT 4; $$
LANGUAGE SQL IMMUTABLE STRICT;

----------------------------------------------------------------
CREATE OR REPLACE FUNCTION datalength(smallint)
RETURNS integer
AS $$ SELECT 2; $$
LANGUAGE SQL IMMUTABLE STRICT;

----------------------------------------------------------------
CREATE OR REPLACE FUNCTION datalength(money)
RETURNS integer
AS $$ SELECT 8; $$
LANGUAGE SQL IMMUTABLE STRICT;

----------------------------------------------------------------
CREATE OR REPLACE FUNCTION datalength(text)
RETURNS integer
AS $$ SELECT octet_length($1); $$
LANGUAGE SQL IMMUTABLE STRICT;

----------------------------------------------------------------
CREATE OR REPLACE FUNCTION datalength(date)
RETURNS integer
AS $$ SELECT 4; $$
LANGUAGE SQL IMMUTABLE STRICT;

----------------------------------------------------------------
CREATE OR REPLACE FUNCTION datalength(time)
RETURNS integer
AS $$ SELECT 8; $$
LANGUAGE SQL IMMUTABLE STRICT;

----------------------------------------------------------------
CREATE OR REPLACE FUNCTION datalength(time with time zone)
RETURNS integer
AS $$ SELECT 12; $$
LANGUAGE SQL IMMUTABLE STRICT;

----------------------------------------------------------------
CREATE OR REPLACE FUNCTION datalength(timestamp)
RETURNS integer
AS $$ SELECT 8; $$
LANGUAGE SQL IMMUTABLE STRICT;

----------------------------------------------------------------
CREATE OR REPLACE FUNCTION datalength(timestamp with time zone)
RETURNS integer
AS $$ SELECT 8; $$
LANGUAGE SQL IMMUTABLE STRICT;

----------------------------------------------------------------
CREATE OR REPLACE FUNCTION datalength(interval)
RETURNS integer
AS $$ SELECT 12; $$
LANGUAGE SQL IMMUTABLE STRICT;

----------------------------------------------------------------
CREATE OR REPLACE FUNCTION datalength(uuid)
RETURNS integer
AS $$ SELECT pg_column_size($1); $$
LANGUAGE SQL IMMUTABLE STRICT;

----------------------------------------------------------------
CREATE OR REPLACE FUNCTION datalength(xml)
RETURNS integer
AS $$ SELECT pg_column_size($1); $$
LANGUAGE SQL IMMUTABLE STRICT;

----------------------------------------------------------------
CREATE OR REPLACE FUNCTION datalength(bytea)
RETURNS integer
AS $$ SELECT octet_length($1); $$
LANGUAGE SQL IMMUTABLE STRICT;

----------------------------------------------------------------
CREATE OR REPLACE FUNCTION datalength(boolean)
RETURNS integer
AS $$ SELECT 1; $$
LANGUAGE SQL IMMUTABLE STRICT;

----------------------------------------------------------------
CREATE OR REPLACE FUNCTION datalength(bit varying)
RETURNS integer
AS $$ SELECT octet_length($1); $$
LANGUAGE SQL IMMUTABLE STRICT;


--<Common function>---------------------------------------------
--
-- Convert from 'VisualStudio QueryBuilder' format to the timestamp format
--
CREATE OR REPLACE FUNCTION vs_format_to_timestamp(varchar)
RETURNS timestamp(3)
AS $$
DECLARE
	tm_val timestamp(3);
	match_val varchar;
BEGIN
	SELECT regexp_matches( TRIM($1), '\d{1,2}/\d{1,2}/\d{4}\s+\d{1,2}:\d{1,2}:\d{1,2}\s+[AP]M' ) INTO match_val;
	IF match_val IS NOT NULL THEN
		--
		-- Format type of "MM/DD/YYYY hh:mm:ss XX"
		--
		SELECT to_timestamp( TRIM($1) ,'MM/DD/YYYY HH12:MI:SS AM' ) INTO tm_val;
		RETURN tm_val;
	END IF;

	SELECT regexp_matches( TRIM($1), '\d{1,2}/\d{1,2}/\d{4}' ) INTO match_val;
	IF match_val IS NOT NULL THEN
		--
		-- Format type of "MM/DD/YYYY"
		--
		SELECT to_timestamp( TRIM($1) ,'MM/DD/YYYY' ) INTO tm_val;
		RETURN tm_val;
	END IF;

	--
	-- Format type of timestamp
	--
	RETURN CAST($1 AS timestamp(3));

END; $$
LANGUAGE PLPGSQL IMMUTABLE STRICT;


--<Common function>---------------------------------------------
--
-- Convert from character strings to the timestamp format
--
CREATE OR REPLACE FUNCTION varchar_to_timestamp(varchar)
RETURNS timestamp(3)
AS $$ 
DECLARE
	position_num integer;
BEGIN
	IF $1 = '' THEN
		RETURN CAST( '1900-01-01 00:00:00' AS timestamp(3) );
	END IF;

	SELECT position(':' IN LTRIM($1)) INTO position_num;
	IF 2 <= position_num AND position_num <=3 THEN
		RETURN CAST('1900-01-01 ' ||  $1 AS timestamp(3) );
	END IF;

	RETURN vs_format_to_timestamp($1);

END; $$ 
LANGUAGE PLPGSQL IMMUTABLE STRICT;


--<DATEADD>-----------------------------------------------------
CREATE OR REPLACE FUNCTION dateadd(varchar,double precision,date)
RETURNS date
AS $$
DECLARE
	retval date;
	number integer;
	item varchar;
BEGIN
	IF $1 IS NULL THEN
		RAISE EXCEPTION 'syntax error at or near "%"', $1;
	ELSIF $2 IS NULL OR $3 IS NULL THEN
		RETURN NULL;
	END IF;
	SELECT CAST(TRUNC($2) AS integer) INTO number;

	SELECT LOWER($1) INTO item;
	CASE item
		WHEN 'year', 'yyyy', 'yy' THEN
			SELECT CAST($3 AS date) + CAST(CAST(number AS varchar) || 'year' AS INTERVAL) INTO retval;
		WHEN 'quarter', 'qq', 'q' THEN
			SELECT CAST($3 AS date) + CAST(CAST(number*3 AS varchar) || 'month' AS INTERVAL) INTO retval;
		WHEN 'month', 'mm', 'm' THEN
			SELECT CAST($3 AS date) + CAST(CAST(number AS varchar) || 'month' AS INTERVAL) INTO retval;
		WHEN 'dayofyear', 'dy', 'y', 'day', 'dd', 'd', 'weekday', 'dw', 'w' THEN
			SELECT CAST($3 AS date) + CAST(CAST(number AS varchar) || 'day' AS INTERVAL) INTO retval;
		WHEN 'week', 'wk', 'ww' THEN
			SELECT CAST($3 AS date) + CAST(CAST(number AS varchar) || 'week' AS INTERVAL) INTO retval;
		ELSE
			RAISE EXCEPTION 'date units "%" not recognized', $1;
	END CASE;
	RETURN retval;
END; $$
LANGUAGE PLPGSQL IMMUTABLE CALLED ON NULL INPUT;

----------------------------------------------------------------
CREATE OR REPLACE FUNCTION dateadd(varchar,double precision,timestamp with time zone)
RETURNS timestamp(3) with time zone
AS $$ 
DECLARE
	retval timestamp(3) with time zone;
	number integer;
	item varchar;
BEGIN
	IF $1 IS NULL THEN
		RAISE EXCEPTION 'syntax error at or near "%"', $1;
	ELSIF $2 IS NULL OR $3 IS NULL THEN
		RETURN NULL;
	END IF;
	SELECT CAST(TRUNC($2) AS integer) INTO number;

	SELECT LOWER($1) INTO item;
	CASE item
		WHEN 'year', 'yyyy', 'yy' THEN
			SELECT CAST($3 AS timestamp with time zone) + CAST(CAST(number AS varchar) || 'year' AS INTERVAL) INTO retval;
		WHEN 'quarter', 'qq', 'q' THEN
			SELECT CAST($3 AS timestamp with time zone) + CAST(CAST(number*3 AS varchar) || 'month' AS INTERVAL) INTO retval;
		WHEN 'month', 'mm', 'm' THEN
			SELECT CAST($3 AS timestamp with time zone) + CAST(CAST(number AS varchar) || 'month' AS INTERVAL) INTO retval;
		WHEN 'dayofyear', 'dy', 'y', 'day', 'dd', 'd', 'weekday', 'dw', 'w' THEN
			SELECT CAST($3 AS timestamp with time zone) + CAST(CAST(number AS varchar) || 'day' AS INTERVAL) INTO retval;
		WHEN 'week', 'wk', 'ww' THEN
			SELECT CAST($3 AS timestamp with time zone) + CAST(CAST(number AS varchar) || 'week' AS INTERVAL) INTO retval;
		WHEN 'hour', 'hh' THEN
			SELECT CAST($3 AS timestamp with time zone) + CAST(CAST(number AS varchar) || 'hour' AS INTERVAL) INTO retval;
		WHEN 'minute', 'mi', 'n' THEN
			SELECT CAST($3 AS timestamp with time zone) + CAST(CAST(number AS varchar) || 'minute' AS INTERVAL) INTO retval;
		WHEN 'second', 'ss', 's' THEN
			SELECT CAST($3 AS timestamp with time zone) + CAST(CAST(number AS varchar) || 'second' AS INTERVAL) INTO retval;
		WHEN 'millisecond', 'ms' THEN
			SELECT CAST($3 AS timestamp with time zone) + CAST(CAST(number AS varchar) || 'millisecond' AS INTERVAL) INTO retval;
		WHEN 'microsecond', 'mcs' THEN
			SELECT CAST($3 AS timestamp with time zone) + CAST(CAST(number AS varchar) || 'microsecond' AS INTERVAL) INTO retval;
		WHEN 'nanosecond', 'ns' THEN
			SELECT CAST($3 AS timestamp with time zone) + CAST(CAST(number/1000 AS varchar) || 'microsecond' AS INTERVAL) INTO retval;
		ELSE
			RAISE EXCEPTION 'timestamp with time zone units "%" not recognized', $1;
	END CASE;
	RETURN retval;
END; $$
LANGUAGE PLPGSQL IMMUTABLE CALLED ON NULL INPUT;

----------------------------------------------------------------
CREATE OR REPLACE FUNCTION dateadd(varchar,double precision,timestamp)
RETURNS timestamp
AS $$ 
DECLARE
	retval timestamp;
BEGIN
	IF $2 IS NULL OR $3 IS NULL THEN
		RETURN NULL;
	END IF;

	SELECT dateadd($1, CAST(TRUNC($2) AS integer),
						CAST($3 AS timestamp with time zone)) INTO retval;
	RETURN CAST(retval AS timestamp);
END; $$
LANGUAGE PLPGSQL IMMUTABLE CALLED ON NULL INPUT;

----------------------------------------------------------------
CREATE OR REPLACE FUNCTION dateadd(varchar,double precision,time with time zone)
RETURNS time with time zone
AS $$ 
DECLARE
	retval time with time zone;
	number integer;
	item varchar;
BEGIN
	IF $1 IS NULL THEN
		RAISE EXCEPTION 'syntax error at or near "%"', $1;
	ELSIF $2 IS NULL OR $3 IS NULL THEN
		RETURN NULL;
	END IF;
	SELECT CAST(TRUNC($2) AS integer) INTO number;

	SELECT LOWER($1) INTO item;
	CASE item
		WHEN 'hour', 'hh' THEN
			SELECT CAST($3 AS time) + CAST(CAST(number AS varchar) || 'hour' AS INTERVAL) INTO retval;
		WHEN 'minute', 'mi', 'n' THEN
			SELECT CAST($3 AS time) + CAST(CAST(number AS varchar) || 'minute' AS INTERVAL) INTO retval;
		WHEN 'second', 'ss', 's' THEN
			SELECT CAST($3 AS time) + CAST(CAST(number AS varchar) || 'second' AS INTERVAL) INTO retval;
		WHEN 'millisecond', 'ms' THEN
			SELECT CAST($3 AS time) + CAST(CAST(number AS varchar) || 'millisecond' AS INTERVAL) INTO retval;
		WHEN 'microsecond', 'mcs' THEN
			SELECT CAST($3 AS time) + CAST(CAST(number AS varchar) || 'microsecond' AS INTERVAL) INTO retval;
		WHEN 'nanosecond', 'ns' THEN
			SELECT CAST($3 AS time) + CAST(CAST(number/1000 AS varchar) || 'microsecond' AS INTERVAL) INTO retval;
		ELSE
			RAISE EXCEPTION 'time units "%" not recognized', $1;
	END CASE;
	RETURN retval;
END; $$
LANGUAGE PLPGSQL IMMUTABLE CALLED ON NULL INPUT;

----------------------------------------------------------------
CREATE OR REPLACE FUNCTION dateadd(varchar,double precision,time)
RETURNS time
AS $$ 
DECLARE
	retval time;
BEGIN
	IF $2 IS NULL OR $3 IS NULL THEN
		RETURN NULL;
	END IF;

	SELECT dateadd($1, CAST(TRUNC($2) AS integer),
						cast($3 as time with time zone)) INTO retval;
	RETURN CAST(retval AS time);
END; $$
LANGUAGE PLPGSQL IMMUTABLE CALLED ON NULL INPUT;

----------------------------------------------------------------
CREATE OR REPLACE FUNCTION dateadd(varchar,double precision,varchar)
RETURNS timestamp(3)
AS $$ 
DECLARE
	retval timestamp(3);
BEGIN
	IF $2 IS NULL OR $3 IS NULL THEN
		RETURN NULL;
	END IF;

	SELECT dateadd($1, CAST(trunc($2) AS integer),
						varchar_to_timestamp($3)) INTO retval;
	RETURN retval;
END; $$
LANGUAGE PLPGSQL IMMUTABLE CALLED ON NULL INPUT;

----------------------------------------------------------------
CREATE OR REPLACE FUNCTION dateadd(varchar,money,date)
RETURNS date
AS $$ SELECT dateadd($1, CAST($2 AS numeric), $3); $$
LANGUAGE SQL IMMUTABLE CALLED ON NULL INPUT;

----------------------------------------------------------------
CREATE OR REPLACE FUNCTION dateadd(varchar,money,timestamp)
RETURNS timestamp
AS $$ SELECT dateadd($1, CAST($2 AS numeric), $3); $$
LANGUAGE SQL IMMUTABLE CALLED ON NULL INPUT;

----------------------------------------------------------------
CREATE OR REPLACE FUNCTION dateadd(varchar,money,timestamp with time zone)
RETURNS timestamp with time zone
AS $$ SELECT dateadd($1, CAST($2 AS numeric), $3); $$
LANGUAGE SQL IMMUTABLE CALLED ON NULL INPUT;

----------------------------------------------------------------
CREATE OR REPLACE FUNCTION dateadd(varchar,money,time)
RETURNS time
AS $$ SELECT dateadd($1, CAST($2 AS numeric), $3); $$
LANGUAGE SQL IMMUTABLE CALLED ON NULL INPUT;

----------------------------------------------------------------
CREATE OR REPLACE FUNCTION dateadd(varchar,money,time with time zone)
RETURNS time with time zone
AS $$ SELECT dateadd($1, CAST($2 AS numeric), $3); $$
LANGUAGE SQL IMMUTABLE CALLED ON NULL INPUT;

----------------------------------------------------------------
CREATE OR REPLACE FUNCTION dateadd(varchar,money,varchar)
RETURNS timestamp(3)
AS $$ SELECT dateadd($1, CAST($2 AS numeric), $3); $$
LANGUAGE SQL IMMUTABLE CALLED ON NULL INPUT;


--<DATEDIFF>----------------------------------------------------
CREATE OR REPLACE FUNCTION datediff(varchar,timestamp with time zone,timestamp with time zone)
RETURNS integer
AS $$
DECLARE
	secondparday integer;
	minuteparday integer;
	secondparhour integer;
	from_date ALIAS FOR $2;
	to_date ALIAS FOR $3;
	retval integer;
	res_feb date;
	last_day integer;
	index integer;
	res_year integer;
	res_day integer;
	res_second numeric;
	s_year integer;
	e_year integer;
	s_month integer;
	e_month integer;
	s_day integer;
	e_day integer;
	s_hour integer;
	e_hour integer;
	s_quarter integer;
	e_quarter integer;
	s_minute integer;
	e_minute integer;
	s_second numeric;
	e_second numeric;
	day1 integer;
	day2 integer;
	day3 integer;
	hour1 integer;
	hour2 integer;
	hour3 integer;
	minute1 integer;
	minute2 integer;
	minute3 integer;
	second1 numeric;
	second2 numeric;
	second3 numeric;
	item varchar;
	inversion integer;
BEGIN
	secondparday = 24*60*60;
	minuteparday = 24*60;
	secondparhour = 60*60;

	IF $1 IS NULL THEN
		RAISE EXCEPTION 'syntax error at or near "%"', $1;
	ELSIF from_date IS NULL OR to_date IS NULL THEN
		RETURN NULL;
	END IF;

	SELECT LOWER($1) INTO item;
	CASE item
		WHEN 'year','yy','yyyy','month','mm','m','quarter','qq','q' THEN
			-- year
			SELECT extract(year FROM from_date) INTO s_year;
			SELECT extract(year FROM to_date) INTO e_year;
			res_year = e_year - s_year;
			IF item = 'year' OR item = 'yy' OR item = 'yyyy' THEN
				RETURN res_year;
			END IF;

			-- month
			SELECT extract(month FROM from_date) INTO s_month;
			SELECT extract(month FROM to_date) INTO e_month;
			IF res_year = 0 THEN
				retval = e_month - s_month;
			ELSIF res_year > 0 THEN
				retval = 12 - s_month + e_month;
				IF res_year > 1 THEN
					retval = retval + 12 * (res_year - 1);
				END IF;
			ELSE
				retval = 12 - e_month + s_month;
				IF res_year < -1 THEN
					retval = retval + 12 * ((res_year * -1) - 1);
				END IF;
				retval = retval * -1;
			END IF;
			IF item = 'month' OR item = 'mm' OR item = 'm' THEN
				RETURN retval;
			END IF;

			-- quarter
			IF s_month >= 1 AND s_month <= 3 THEN
				s_quarter = 1;
			ELSIF s_month >= 4 AND s_month <= 6 THEN
				s_quarter = 2;
			ELSIF s_month >= 7 AND s_month <= 9 THEN
				s_quarter = 3;
			ELSE
				s_quarter = 4;
			END IF;

			IF e_month >= 1 AND e_month <= 3 THEN
				e_quarter = 1;
			ELSIF e_month >= 4 AND e_month <= 6 THEN
				e_quarter = 2;
			ELSIF e_month >= 7 AND e_month <= 9 THEN
				e_quarter = 3;
			ELSE
				e_quarter = 4;
			END IF;

			retval = e_quarter + (res_year * 4) - s_quarter;
			RETURN retval;

		WHEN 'day','dd','d','dayofyear','dy','y',
			 'day','dd','d','week','wk','ww','hour','hh',
			 'minute','mi','n','second','ss','s','millisecond','ms',
			 'microsecond','mcs','nanosecond','ns' THEN
			-- day
			IF from_date <= to_date THEN
				SELECT extract(year FROM from_date) INTO s_year;
				SELECT extract(year FROM to_date) INTO e_year;
				SELECT extract(month FROM from_date) INTO s_month;
				SELECT extract(month FROM to_date) INTO e_month;
				SELECT extract(day FROM from_date) INTO s_day;
				SELECT extract(day FROM to_date) INTO e_day;
				inversion = 1;
			ELSE
				SELECT extract(year FROM to_date) INTO s_year;
				SELECT extract(year FROM from_date) INTO e_year;
				SELECT extract(month FROM to_date) INTO s_month;
				SELECT extract(month FROM from_date) INTO e_month;
				SELECT extract(day FROM to_date) INTO s_day;
				SELECT extract(day FROM from_date) INTO e_day;
				inversion = -1;
			END IF;

			day1 = 0;
			index = 1;
			WHILE index <= s_month - 1 LOOP
				CASE index
					WHEN 1,3,5,7,8,10,12 THEN
						day1 = day1 + 31;
					WHEN 4,6,9,11 THEN
						day1 = day1 + 30;
					WHEN 2 THEN
						SELECT date(date_trunc('MONTH',CAST(s_year||'-02-01' AS date)) + 
						INTERVAL'1 MONTH' - INTERVAL'1 DAY') INTO res_feb;
						SELECT extract(day FROM res_feb) INTO last_day;
						day1 = day1 + last_day;
				END CASE;
				index = index + 1;
			END LOOP;
			day1 = day1 + s_day;

			day2 = 0;
			index = s_year;
			WHILE index <= e_year - 1 LOOP
				day2 = day2 + 365;

				SELECT date(date_trunc('MONTH',CAST(index||'-02-01' AS date)) + 
								INTERVAL'1 MONTH' - INTERVAL'1 DAY') INTO res_feb;
				SELECT extract(day FROM res_feb) INTO last_day;

				IF last_day = 29 THEN
					day2 = day2 + 1;
				END IF;
				index = index + 1;
			END LOOP;

			day3 = 0;
			index = 1;
			WHILE index <= e_month - 1 LOOP
				CASE index
					WHEN 1,3,5,7,8,10,12 THEN
						day3 = day3 + 31;
					WHEN 4,6,9,11 THEN
						day3 = day3 + 30;
					WHEN 2 THEN
						SELECT date(date_trunc('MONTH',CAST(e_year||'-02-01' AS date)) +
						INTERVAL'1 MONTH' - INTERVAL'1 DAY') INTO res_feb;
						SELECT extract(day FROM res_feb) INTO last_day;
						day3 = day3 + last_day;
				END CASE;
				index = index + 1;
			END LOOP;
			day3 = day3 + e_day;

			res_day = day2 + day3 - day1;
			IF item = 'day' OR item = 'dd' OR item = 'd' OR item = 'dayofyear' OR item = 'y' OR item = 'dy'THEN
				RETURN res_day * inversion;
			END IF;

			-- week
			IF item = 'week' OR item = 'wk' OR item = 'ww'THEN
				retval = res_day / 7;
				RETURN retval;
			END IF;
			-- hour
			IF item = 'hour' OR item = 'hh' THEN
				IF from_date <= to_date THEN
					SELECT extract(hour FROM from_date) INTO s_hour;
					SELECT extract(hour FROM to_date) INTO e_hour;
				ELSE
					SELECT extract(hour FROM to_date) INTO s_hour;
					SELECT extract(hour FROM from_date) INTO e_hour;
				END IF;	

				hour1 = 24- s_hour;
				hour2 = (res_day - 1) * 24;
				hour3 = 24 - (24 - e_hour);
				retval = hour2 + hour3 + hour1;
				RETURN retval * inversion;
			END IF;

			-- monute
			IF item = 'minute' OR item = 'mi' OR item = 'n' THEN
				IF from_date <= to_date THEN
					SELECT extract(hour FROM from_date) INTO s_hour;
					SELECT extract(hour FROM to_date) INTO e_hour;
					SELECT extract(minute FROM from_date) INTO s_minute;
					SELECT extract(minute FROM to_date) INTO e_minute;
				ELSE
					SELECT extract(hour FROM to_date) INTO s_hour;
					SELECT extract(hour FROM from_date) INTO e_hour;
					SELECT extract(minute FROM to_date) INTO s_minute;
					SELECT extract(minute FROM from_date) INTO e_minute;
				END IF;	
				minute1 = minuteparday - ((s_hour * 60) + s_minute);
				minute2 = (res_day - 1) * minuteparday;
				minute3 = minuteparday - (((23 - e_hour) * 60) + (60 - e_minute));
				retval = minute2 + minute3 + minute1;
				RETURN retval * inversion;
			END IF;

			-- second
			CASE item
				WHEN 'second','ss','s','millisecond','ms','microsecond','mcs',
					 'nanosecond','ns' THEN

					IF from_date <= to_date THEN
						SELECT extract(hour FROM from_date) INTO s_hour;
						SELECT extract(hour FROM to_date) INTO e_hour;
						SELECT extract(minute FROM from_date) INTO s_minute;
						SELECT extract(minute FROM to_date) INTO e_minute;
						SELECT extract(second FROM from_date) INTO s_second;
						SELECT extract(second FROM to_date) INTO e_second;
					ELSE
						SELECT extract(hour FROM to_date) INTO s_hour;
						SELECT extract(hour FROM from_date) INTO e_hour;
						SELECT extract(minute FROM to_date) INTO s_minute;
						SELECT extract(minute FROM from_date) INTO e_minute;
						SELECT extract(second FROM to_date) INTO s_second;
						SELECT extract(second FROM from_date) INTO e_second;
					END IF;	
					second1 = secondparday - ((s_hour * secondparhour) + (s_minute * 60) + s_second);
					second2 = (res_day - 1) * secondparday;
					second3 = secondparday - (((23 - e_hour) * secondparhour) + ((59 - e_minute) * 60) + (60 - e_second));
					res_second = second2 + second3 + second1;

					IF item = 'second' OR item = 'ss' OR item = 's' THEN
						SELECT CAST(trunc(res_second + .9, 0) AS integer) INTO retval;
						RETURN retval * inversion;
					END IF;

					-- milli second
					IF item = 'millisecond' OR item = 'ms' THEN
						SELECT CAST(trunc(res_second * 1000 + .9, 0) AS integer) INTO retval;
						RETURN retval * inversion;
					END IF;

					-- micro second
					IF item = 'microsecond' OR item = 'mcs' THEN
						SELECT CAST(trunc(res_second * 1000 * 1000 + .9, 0) AS integer) INTO retval;
						RETURN retval * inversion;
					END IF;

					-- nano second
					IF item = 'nanosecond' OR item = 'ns' THEN
						SELECT CAST(trunc(res_second * 1000 * 1000 * 1000 + .9, 0) AS integer) INTO retval;
						RETURN retval * inversion;
					END IF;
			END CASE;
		ELSE
			RAISE EXCEPTION 'timestamp units "%" not recognized', $1;
	END CASE;

END; $$
LANGUAGE PLPGSQL IMMUTABLE CALLED ON NULL INPUT;

----------------------------------------------------------------
CREATE OR REPLACE FUNCTION datediff(varchar,timestamp with time zone,varchar)
RETURNS integer
AS $$
DECLARE
	retval integer;
BEGIN
	IF $2 IS NULL OR $3 IS NULL THEN
		RETURN NULL;
	END IF;

	SELECT datediff($1,$2,varchar_to_timestamp($3)) INTO retval;
	RETURN retval;
END; $$
LANGUAGE PLPGSQL IMMUTABLE CALLED ON NULL INPUT;

----------------------------------------------------------------
CREATE OR REPLACE FUNCTION datediff(varchar,timestamp with time zone,time with time zone)
RETURNS integer
AS $$
DECLARE
	retval integer;
BEGIN
	IF $2 IS NULL OR $3 IS NULL THEN
		RETURN NULL;
	END IF;

	SELECT datediff($1,$2,CAST('1900-01-01 '||$3 AS timestamp)) INTO retval;
	RETURN retval;
END; $$
LANGUAGE PLPGSQL IMMUTABLE CALLED ON NULL INPUT;

----------------------------------------------------------------
CREATE OR REPLACE FUNCTION datediff(varchar,varchar,timestamp with time zone)
RETURNS integer
AS $$
DECLARE
	retval integer;
BEGIN
	IF $2 IS NULL OR $3 IS NULL THEN
		RETURN NULL;
	END IF;

	SELECT datediff($1,varchar_to_timestamp($2),$3) INTO retval;
	RETURN retval;
END; $$
LANGUAGE PLPGSQL IMMUTABLE CALLED ON NULL INPUT;

----------------------------------------------------------------
CREATE OR REPLACE FUNCTION datediff(varchar,varchar,varchar)
RETURNS integer
AS $$
DECLARE
	retval integer;
BEGIN
	IF $2 IS NULL OR $3 IS NULL THEN
		RETURN NULL;
	END IF;

	SELECT datediff($1,varchar_to_timestamp($2),
						varchar_to_timestamp($3)) INTO retval;
	RETURN retval;
END; $$
LANGUAGE PLPGSQL IMMUTABLE CALLED ON NULL INPUT;

----------------------------------------------------------------
CREATE OR REPLACE FUNCTION datediff(varchar,varchar,time with time zone)
RETURNS integer
AS $$
DECLARE
	retval integer;
BEGIN
	IF $2 IS NULL OR $3 IS NULL THEN
		RETURN NULL;
	END IF;

	SELECT datediff($1,varchar_to_timestamp($2),
						CAST('1900-01-01 '||$3 AS timestamp)) INTO retval;
	RETURN retval;
END; $$
LANGUAGE PLPGSQL IMMUTABLE CALLED ON NULL INPUT;

----------------------------------------------------------------
CREATE OR REPLACE FUNCTION datediff(varchar,time with time zone,timestamp with time zone)
RETURNS integer
AS $$
DECLARE
	retval integer;
BEGIN
	IF $2 IS NULL OR $3 IS NULL THEN
		RETURN NULL;
	END IF;

	SELECT datediff($1,CAST('1900-01-01 '||$2 AS timestamp),$3) INTO retval;
	RETURN retval;
END; $$
LANGUAGE PLPGSQL IMMUTABLE CALLED ON NULL INPUT;

----------------------------------------------------------------
CREATE OR REPLACE FUNCTION datediff(varchar,time with time zone,varchar)
RETURNS integer
AS $$
DECLARE
	retval integer;
BEGIN
	IF $2 IS NULL OR $3 IS NULL THEN
	 RETURN NULL;
	END IF;

	SELECT datediff($1,CAST('1900-01-01 '||$2 AS timestamp),
						varchar_to_timestamp($3)) INTO retval;
	RETURN retval;
END; $$
LANGUAGE PLPGSQL IMMUTABLE CALLED ON NULL INPUT;

----------------------------------------------------------------
CREATE OR REPLACE FUNCTION datediff(varchar,time with time zone,time with time zone)
RETURNS integer
AS $$
DECLARE
	retval integer;
BEGIN
	IF $2 IS NULL OR $3 IS NULL THEN
		RETURN NULL;
	END IF;

	SELECT datediff($1,CAST('1900-01-01 '||$2 AS timestamp),
						CAST('1900-01-01 '||$3 AS timestamp)) INTO retval;
	RETURN retval;
END; $$
LANGUAGE PLPGSQL IMMUTABLE CALLED ON NULL INPUT;


--<DATEPART>----------------------------------------------------
CREATE OR REPLACE FUNCTION datepart(varchar,date)
RETURNS	 integer
AS $$
DECLARE
	tmp_val numeric;
	week_val double precision;
	retval integer;
	item varchar;
BEGIN
	IF $1 IS NULL THEN
		RAISE EXCEPTION 'syntax error at or near "%"', $1;
	ELSIF $2 IS NULL THEN
		RETURN NULL;
	END IF;

	SELECT LOWER($1) INTO item;
	CASE item
		WHEN 'year', 'yyyy', 'yy' THEN
			SELECT extract(year FROM $2) INTO retval;
		WHEN 'quarter', 'qq', 'q' THEN
			SELECT extract(quarter FROM $2) INTO retval;
		WHEN 'month', 'mm', 'm' THEN
			SELECT extract(month FROM $2) INTO retval;
		WHEN 'dayofyear', 'dy', 'y' THEN
			SELECT extract(doy FROM $2) INTO retval;
		WHEN 'day', 'dd', 'd' THEN
			SELECT extract(day FROM $2) INTO retval;
		WHEN 'week', 'wk', 'ww' THEN
			-- SQLServer (January 1 starting point)
			--  /PostgreSQL (week starting point with the first Thursday)
			-- In order to cancel a difference
			SELECT extract(year FROM $2) INTO tmp_val;
			SELECT extract(week FROM CAST(tmp_val || '-01-01' AS date))
																INTO week_val;
			IF week_val = 53 THEN
				SELECT extract(week FROM $2) +1 INTO retval;
				IF retval >= 53 THEN
					retval = 1;
				END IF;
			ELSIF week_val = 52 THEN
				SELECT extract(week FROM $2) +2 INTO retval;
				IF retval >= 52 THEN
					retval = 1;
				END IF;
			ELSE
				SELECT extract(week FROM $2) INTO retval;
			END IF;
		WHEN 'weekday', 'dw' THEN
			--It takes +1, in order to cancel the difference in (0) on (1) 
			-- / PostgreSQL Sunday on SQLServer Sunday.
			SELECT extract(dow FROM $2) + 1 INTO retval;
		WHEN 'ISO_WEEK', 'isowk', 'isoww' THEN
			SELECT extract(week FROM $2) INTO retval;
		ELSE
			RAISE EXCEPTION 'date units "%" not recognized', $1;
	END CASE;
	RETURN retval;
END; $$
LANGUAGE PLPGSQL IMMUTABLE CALLED ON NULL INPUT;

----------------------------------------------------------------
CREATE OR REPLACE FUNCTION datepart(varchar,timestamp)
RETURNS	 integer
AS $$
DECLARE
	retval integer;
BEGIN
	IF $1 = 'TZoffset' OR $1 = 'tz' THEN
		RAISE EXCEPTION 'timestamp units "%" not recognized', $1;
	END IF;
	SELECT datepart($1, CAST($2 AS timestamp with time zone)) INTO retval;
	RETURN retval;
END; $$
LANGUAGE PLPGSQL IMMUTABLE CALLED ON NULL INPUT;

----------------------------------------------------------------
CREATE OR REPLACE FUNCTION datepart(varchar,timestamp with time zone)
RETURNS	 integer
AS $$
DECLARE
	tmp_val numeric;
	week_val double precision;
	retval integer;
	item varchar;
BEGIN
	IF $1 IS NULL THEN
		RAISE EXCEPTION 'syntax error at or near "%"', $1;
	ELSIF $2 IS NULL THEN
		RETURN NULL;
	END IF;

	SELECT LOWER($1) INTO item;
	CASE item
		WHEN 'year', 'yyyy', 'yy' THEN
			SELECT extract(year FROM $2) INTO retval;
		WHEN 'quarter', 'qq', 'q' THEN
			SELECT extract(quarter FROM $2) INTO retval;
		WHEN 'month', 'mm', 'm' THEN
			SELECT extract(month FROM $2) INTO retval;
		WHEN 'dayofyear', 'dy', 'y' THEN
			SELECT extract(doy FROM $2) INTO retval;
		WHEN 'day', 'dd', 'd' THEN
			SELECT extract(day FROM $2) INTO retval;
		WHEN 'week', 'wk', 'ww' THEN
			-- SQLServer (January 1 starting point)/PostgreSQL (week starting point with the first Thursday)
			-- In order to cancel a difference 
			SELECT extract(year FROM $2) INTO tmp_val;
			SELECT extract(week FROM CAST(tmp_val || '-01-01' AS date)) INTO week_val;
			IF week_val = 53 THEN
				SELECT extract(week FROM $2) +1 INTO retval;
				IF retval >= 53 THEN
					retval = 1;
				END IF;
			ELSIF week_val = 52 THEN
				SELECT extract(week FROM $2) +2 INTO retval;
				IF retval >= 52 THEN
					retval = 1;
				END IF;
			ELSE
				SELECT extract(week FROM $2) INTO retval;
			END IF;
		WHEN 'weekday', 'dw' THEN
			-- It takes +1, in order to cancel the difference in (0) on (1) 
			-- / PostgreSQL Sunday on SQLServer Sunday. 

			SELECT extract(dow FROM $2) + 1 INTO retval;
		WHEN 'hour', 'hh' THEN
			SELECT extract(hour FROM $2) INTO retval;
		WHEN 'minute', 'mi', 'n' THEN
			SELECT extract(minute FROM $2) INTO retval;
		WHEN 'second', 'ss', 's' THEN
			SELECT extract(second FROM $2) INTO tmp_val;
			SELECT trunc(tmp_val) INTO retval;
		WHEN 'millisecond', 'ms' THEN
			SELECT CAST(extract(milliseconds FROM $2) AS numeric(18,3)) / 1000 INTO tmp_val;
			SELECT trunc((tmp_val - trunc(tmp_val)) * 1000) INTO retval;
		WHEN 'microsecond', 'mcs' THEN
			SELECT CAST(extract(microseconds FROM $2) AS numeric(18,6)) / 1000000 INTO tmp_val;
			SELECT trunc((tmp_val - trunc(tmp_val)) * 1000000) INTO retval;
		WHEN 'nanosecond', 'ns' THEN
			SELECT CAST(extract(microseconds FROM $2) AS numeric(18,6)) / 1000000 INTO tmp_val;
			SELECT trunc((tmp_val - trunc(tmp_val)) * 1000000000) INTO retval;
		WHEN 'TZoffset','tzoffset','tz' THEN
			SELECT extract(timezone FROM $2) / 60 INTO retval;
		WHEN 'ISO_WEEK','iso_week', 'isowk', 'isoww' THEN
			SELECT extract(week FROM $2) INTO retval;
		ELSE
			RAISE EXCEPTION 'timestamp with time zone units "%" not recognized', $1;
	END CASE;
	RETURN retval;
END; $$
LANGUAGE PLPGSQL IMMUTABLE CALLED ON NULL INPUT;

----------------------------------------------------------------
CREATE OR REPLACE FUNCTION datepart(varchar,time)
RETURNS	 integer
AS $$
DECLARE
	retval integer;
BEGIN
	IF $1 = 'TZoffset' OR $1 = 'tz' THEN
		RAISE EXCEPTION 'time units "%" not recognized', $1;
	END IF;
	SELECT datepart($1, CAST($2 AS time with time zone)) INTO retval;
	RETURN retval;
END; $$
LANGUAGE PLPGSQL IMMUTABLE CALLED ON NULL INPUT;

----------------------------------------------------------------
CREATE OR REPLACE FUNCTION datepart(varchar,time with time zone)
RETURNS	 integer
AS $$
DECLARE
	tmp_val double precision;
	week_val double precision;
	retval integer;
	item varchar;
BEGIN
	IF $1 IS NULL THEN
		RAISE EXCEPTION 'syntax error at or near "%"', $1;
	ELSIF $2 IS NULL THEN
		RETURN NULL;
	END IF;
	SELECT LOWER($1) INTO item;
	CASE item
		WHEN 'hour', 'hh' THEN
			SELECT extract(hour FROM $2) INTO retval;
		WHEN 'minute', 'mi', 'n' THEN
			SELECT extract(minute FROM $2) INTO retval;
		WHEN 'second', 'ss', 's' THEN
			SELECT extract(second FROM $2) INTO tmp_val;
			SELECT trunc(tmp_val) INTO retval;
		WHEN 'millisecond', 'ms' THEN
			SELECT CAST(extract(milliseconds FROM $2) AS numeric(18,3)) / 1000 INTO tmp_val;
			SELECT trunc((tmp_val - trunc(tmp_val)) * 1000) INTO retval;
		WHEN 'microsecond', 'mcs' THEN
			SELECT CAST(extract(microseconds FROM $2) AS numeric(18,6)) / 1000000 INTO tmp_val;
			SELECT trunc((tmp_val - trunc(tmp_val)) * 1000000) INTO retval;
		WHEN 'nanosecond', 'ns' THEN
			SELECT CAST(extract(microseconds FROM $2) AS numeric(18,6)) / 1000000 INTO tmp_val;
			SELECT trunc((tmp_val - trunc(tmp_val)) * 1000000000) INTO retval;
		WHEN 'TZoffset', 'tz' THEN
			SELECT extract(timezone FROM $2) / 60 INTO retval;
		ELSE
			RAISE EXCEPTION 'time with time zone units "%" not recognized', $1;
	END CASE;
	RETURN retval;
END; $$
LANGUAGE PLPGSQL IMMUTABLE CALLED ON NULL INPUT;

----------------------------------------------------------------
CREATE OR REPLACE FUNCTION datepart(varchar,varchar)
RETURNS	 integer
AS $$
DECLARE
	retval integer;
BEGIN
	IF $1 IS NULL THEN
		RAISE EXCEPTION 'syntax error at or near "%"', $1;
	ELSIF $2 IS NULL THEN
		RETURN NULL;
	END IF;
	IF $1 = 'TZoffset' OR $1 = 'tz' THEN
		RAISE EXCEPTION 'timestamp units "%" not recognized', $1;
	END IF;

	SELECT datepart($1,
					CAST(varchar_to_timestamp($2) AS timestamp with time zone))
					INTO retval;
	RETURN retval;
END; $$
LANGUAGE PLPGSQL IMMUTABLE CALLED ON NULL INPUT;


--<DAY>---------------------------------------------------------
CREATE OR REPLACE FUNCTION day(timestamp with time zone)
RETURNS double precision
AS $$ SELECT extract(DAY FROM date($1)); $$
LANGUAGE SQL IMMUTABLE STRICT;

----------------------------------------------------------------
CREATE OR REPLACE FUNCTION day(varchar)
RETURNS double precision
AS $$ SELECT extract(DAY FROM varchar_to_timestamp($1)); $$ 
LANGUAGE SQL IMMUTABLE STRICT;

----------------------------------------------------------------
CREATE OR REPLACE FUNCTION day(double precision)
RETURNS double precision
AS $$ 
SELECT extract (DAY FROM timestamp '1900-01-01'
					+ CAST(CAST($1*60*60*24 AS varchar) AS interval)); $$
LANGUAGE SQL IMMUTABLE STRICT;


--<MONTH>-------------------------------------------------------
CREATE OR REPLACE FUNCTION month(timestamp with time zone)
RETURNS double precision
AS $$ SELECT extract(MONTH FROM date($1)); $$
LANGUAGE SQL IMMUTABLE STRICT;

----------------------------------------------------------------
CREATE OR REPLACE FUNCTION month(varchar)
RETURNS double precision
AS $$ SELECT extract(MONTH FROM varchar_to_timestamp($1)); $$ 
LANGUAGE SQL IMMUTABLE STRICT;

----------------------------------------------------------------
CREATE OR REPLACE FUNCTION month(double precision)
RETURNS double precision
AS $$ 
SELECT extract (MONTH FROM timestamp '1900-01-01'
					+ CAST(CAST($1*60*60*24 AS varchar) AS interval)); $$
LANGUAGE SQL IMMUTABLE STRICT;


--<YEAR>--------------------------------------------------------
CREATE OR REPLACE FUNCTION year(timestamp with time zone)
RETURNS double precision
AS $$ SELECT extract(YEAR FROM date($1)); $$
LANGUAGE SQL IMMUTABLE STRICT;

----------------------------------------------------------------
CREATE OR REPLACE FUNCTION year(varchar)
RETURNS double precision
AS $$ SELECT extract(YEAR FROM varchar_to_timestamp($1)); $$ 
LANGUAGE SQL IMMUTABLE STRICT;

----------------------------------------------------------------
CREATE OR REPLACE FUNCTION year(double precision)
RETURNS double precision
AS $$ 
SELECT extract (YEAR FROM timestamp '1900-01-01'
					+ CAST(CAST($1*60*60*24 AS varchar) AS interval)); $$
LANGUAGE SQL IMMUTABLE STRICT;


--<EOMONTH>-----------------------------------------------------
CREATE OR REPLACE FUNCTION eomonth(timestamp with time zone, double precision)
RETURNS date
AS $$
DECLARE
	retval date;
	month_to_add ALIAS FOR $2;
BEGIN
	IF $1 IS NULL THEN
		RETURN NULL;
	END IF;
	IF month_to_add IS NULL THEN
		month_to_add = 0;
	END IF;
	SELECT date(date_trunc('MONTH',$1)
				+ CAST(trunc(month_to_add) || ' MONTH' AS INTERVAL)
				+'1 MONTH'
				- INTERVAL '1 DAY') INTO retval;
	RETURN retval;
END; $$
LANGUAGE PLPGSQL IMMUTABLE CALLED ON NULL INPUT;

----------------------------------------------------------------
CREATE OR REPLACE FUNCTION eomonth(varchar, double precision)
RETURNS date
AS $$ 
DECLARE
	position_num integer;
	retval date;
BEGIN
	SELECT position(':' IN LTRIM($1)) INTO position_num;
	IF $1 IS NULL THEN
		RETURN NULL;
	ELSIF $1 = '' THEN
		SELECT eomonth(date('1900-01-31'), $2) INTO retval;
		RETURN retval;
	ELSIF 1 <= position_num AND position_num <=3 THEN
		PERFORM CAST($1 AS time);
		SELECT eomonth(date('1900-01-31'), $2) INTO retval;
		RETURN retval;
	END IF;

	SELECT eomonth(date($1), $2) INTO retval;
	RETURN retval;
END; $$ 
LANGUAGE PLPGSQL IMMUTABLE CALLED ON NULL INPUT;

----------------------------------------------------------------
CREATE OR REPLACE FUNCTION eomonth(timestamp with time zone)
RETURNS date
AS $$ SELECT date(date_trunc('MONTH',$1) + INTERVAL '1 MONTH' - INTERVAL '1 DAY'); $$
LANGUAGE SQL IMMUTABLE STRICT;

----------------------------------------------------------------
CREATE OR REPLACE FUNCTION eomonth(varchar)
RETURNS date
AS $$ SELECT eomonth($1,0); $$
LANGUAGE SQL IMMUTABLE STRICT;

----------------------------------------------------------------
CREATE OR REPLACE FUNCTION eomonth(timestamp with time zone, money)
RETURNS date
AS $$
SELECT eomonth($1,cast($2 as numeric)); $$
LANGUAGE SQL IMMUTABLE CALLED ON NULL INPUT;

----------------------------------------------------------------
CREATE OR REPLACE FUNCTION eomonth(varchar, money)
RETURNS date
AS $$
SELECT eomonth($1,cast($2 as numeric)); $$
LANGUAGE SQL IMMUTABLE CALLED ON NULL INPUT;


--<GETDATE>-----------------------------------------------------
CREATE OR REPLACE FUNCTION getdate()
RETURNS timestamp
AS $$ SELECT localtimestamp(3); $$
LANGUAGE SQL IMMUTABLE STRICT;


--<SYSDATETIME>-------------------------------------------------
CREATE OR REPLACE FUNCTION sysdatetime()
RETURNS timestamp
AS $$ SELECT localtimestamp; $$
LANGUAGE SQL IMMUTABLE STRICT;


--<	OPERAT!> >--------------------------------------------------
BEGIN;
CREATE TABLE toperatornotgreaterthan
(
  oprname name NOT NULL,
  oprnamespace oid NOT NULL,
  oprowner oid NOT NULL,
  oprkind "char" NOT NULL,
  oprcanmerge boolean NOT NULL,
  oprcanhash boolean NOT NULL,
  oprleft oid NOT NULL,
  oprright oid NOT NULL,
  oprresult oid NOT NULL,
  oprcom oid NOT NULL,
  oprnegate oid NOT NULL,
  oprcode regproc NOT NULL,
  oprrest regproc NOT NULL,
  oprjoin regproc NOT NULL
);
INSERT INTO toperatornotgreaterthan SELECT * FROM pg_catalog.pg_operator WHERE oprname='<=';
UPDATE toperatornotgreaterthan SET oprname='!>';
INSERT INTO pg_catalog.pg_operator SELECT * FROM toperatornotgreaterthan;
DROP TABLE toperatornotgreaterthan;
COMMIT;


--<	OPERAT !< >-------------------------------------------------
BEGIN;
CREATE TABLE toperatornotlessthan
(
  oprname name NOT NULL,
  oprnamespace oid NOT NULL,
  oprowner oid NOT NULL,
  oprkind "char" NOT NULL,
  oprcanmerge boolean NOT NULL,
  oprcanhash boolean NOT NULL,
  oprleft oid NOT NULL,
  oprright oid NOT NULL,
  oprresult oid NOT NULL,
  oprcom oid NOT NULL,
  oprnegate oid NOT NULL,
  oprcode regproc NOT NULL,
  oprrest regproc NOT NULL,
  oprjoin regproc NOT NULL
);
INSERT INTO toperatornotlessthan SELECT * FROM pg_catalog.pg_operator WHERE oprname='>=';
UPDATE toperatornotlessthan SET oprname='!<';
INSERT INTO pg_catalog.pg_operator SELECT * FROM toperatornotlessthan;
DROP TABLE toperatornotlessthan;
COMMIT;


--<	DATE TYPE >-------------------------------------------------
CREATE domain datetime as timestamp(3);
CREATE domain datetime2 as timestamp(6);
CREATE domain smalldatetime as timestamp(0);
CREATE domain uniqueidentifier as uuid;


--<	odbc_time >-------------------------------------------------
CREATE OR REPLACE FUNCTION odbc_time(varchar) 
RETURNS timestamp
AS $$ 
DECLARE
    date_section varchar;
    retval varchar;
BEGIN
    SELECT CURRENT_DATE INTO date_section;
    SELECT date_section||' '||$1 INTO retval;
    RETURN cast(retval as timestamp(3));
END; $$ 
LANGUAGE PLPGSQL IMMUTABLE CALLED ON NULL INPUT;


--<	c >---------------------------------------------------------
CREATE OR REPLACE FUNCTION c(double precision) 
RETURNS time 
AS $$ 
BEGIN
IF $1<=0  THEN
    RETURN localtime(0);
ELSIF $1 IS NULL THEN
    RETURN localtime(3);
ELSIF $1>=6 THEN
    RETURN localtime(6);
ELSE
    RETURN cast(substr(cast(localtime(6) as varchar),1,9+cast(trunc($1) as integer))as time);
END IF;
END; $$ 
LANGUAGE PLPGSQL IMMUTABLE CALLED ON NULL INPUT;


--<	c >---------------------------------------------------------
CREATE OR REPLACE FUNCTION c(money) 
RETURNS time 
AS $$ 
select c(cast($1 as numeric));
$$ 
LANGUAGE SQL IMMUTABLE CALLED ON NULL INPUT;